@@ -6,34 +6,25 @@ can work with any of the supported filesystems.
 
 """
 
-from __future__ import absolute_import
-from __future__ import print_function
-from __future__ import unicode_literals
+from __future__ import absolute_import, print_function, unicode_literals
 
 import abc
+import hashlib
+import itertools
 import os
 import threading
 import time
 import typing
-from functools import partial
-
 from contextlib import closing
-import itertools
+from functools import partial, wraps
+import warnings
 
 import six
 
-from . import copy
-from . import errors
-from . import fsencode
-from . import iotools
-from . import move
-from . import tools
-from . import walk
-from . import wildcard
+from . import copy, errors, fsencode, iotools, move, tools, walk, wildcard
+from .glob import BoundGlobber
 from .mode import validate_open_mode
-from .path import abspath
-from .path import join
-from .path import normpath
+from .path import abspath, join, normpath
 from .time import datetime_to_epoch
 from .walk import Walker
 
@@ -72,6 +63,33 @@ if False:  # typing.TYPE_CHECKING
 __all__ = ["FS"]
 
 
+def _new_name(method, old_name):
+    """Return a method with a deprecation warning."""
+    # Looks suspiciously like a decorator, but isn't!
+
+    @wraps(method)
+    def _method(*args, **kwargs):
+        warnings.warn(
+            "method '{}' has been deprecated, please rename to '{}'".format(
+                old_name, method.__name__
+            ),
+            DeprecationWarning,
+        )
+        return method(*args, **kwargs)
+
+    deprecated_msg = """
+        Note:
+            .. deprecated:: 2.2.0
+                Please use `~{}`
+""".format(
+        method.__name__
+    )
+    if getattr(_method, "__doc__"):
+        _method.__doc__ += deprecated_msg
+
+    return _method
+
+
 @six.add_metaclass(abc.ABCMeta)
 class FS(object):
     """Base class for FS objects.
@@ -108,6 +126,12 @@ class FS(object):
         """
         self.close()
 
+    @property
+    def glob(self):
+        """`~fs.glob.BoundGlobber`: a globber object..
+        """
+        return BoundGlobber(self)
+
     @property
     def walk(self):
         # type: (_F) -> BoundWalker[_F]
@@ -121,12 +145,8 @@ class FS(object):
     # ---------------------------------------------------------------- #
 
     @abc.abstractmethod
-    def getinfo(
-        self,
-        path,  # type: Text
-        namespaces=None,  # type: Optional[Collection[Text]]
-    ):
-        # type: (...) -> Info
+    def getinfo(self, path, namespaces=None):
+        # type: (Text, Optional[Collection[Text]]) -> Info
         """Get information about a resource on a filesystem.
 
         Arguments:
@@ -355,7 +375,7 @@ class FS(object):
 
         Example:
             >>> with OSFS('~/Desktop') as desktop_fs:
-            ...    desktop_fs.settext(
+            ...    desktop_fs.writetext(
             ...        'note.txt',
             ...        "Don't forget to tape Game of Thrones"
             ...    )
@@ -388,7 +408,7 @@ class FS(object):
                 raise errors.DestinationExists(dst_path)
             with closing(self.open(src_path, "rb")) as read_file:
                 # FIXME(@althonos): typing complains because open return IO
-                self.setbinfile(dst_path, read_file)  # type: ignore
+                self.upload(dst_path, read_file)  # type: ignore
 
     def copydir(self, src_path, dst_path, create=False):
         # type: (Text, Text, bool) -> None
@@ -398,7 +418,7 @@ class FS(object):
             src_path (str): Path of source directory.
             dst_path (str): Path to destination directory.
             create (bool): If `True`, then ``dst_path`` will be created
-                if it doesn't exist alreadys (defaults to `False`).
+                if it doesn't exist already (defaults to `False`).
 
         Raises:
             fs.errors.ResourceNotFound: If the ``dst_path``
@@ -559,7 +579,7 @@ class FS(object):
             iter_info = itertools.islice(iter_info, start, end)
         return iter_info
 
-    def getbytes(self, path):
+    def readbytes(self, path):
         # type: (Text) -> bytes
         """Get the contents of a file as bytes.
 
@@ -577,7 +597,9 @@ class FS(object):
             contents = read_file.read()
         return contents
 
-    def getfile(self, path, file, chunk_size=None, **options):
+    getbytes = _new_name(readbytes, "getbytes")
+
+    def download(self, path, file, chunk_size=None, **options):
         # type: (Text, BinaryIO, Optional[int], **Any) -> None
         """Copies a file from the filesystem to a file-like object.
 
@@ -600,14 +622,16 @@ class FS(object):
 
         Example:
             >>> with open('starwars.mov', 'wb') as write_file:
-            ...     my_fs.getfile('/movies/starwars.mov', write_file)
+            ...     my_fs.download('/movies/starwars.mov', write_file)
 
         """
         with self._lock:
             with self.openbin(path, **options) as read_file:
                 tools.copy_file_data(read_file, file, chunk_size=chunk_size)
 
-    def gettext(
+    getfile = _new_name(download, "getfile")
+
+    def readtext(
         self,
         path,  # type: Text
         encoding=None,  # type: Optional[Text]
@@ -639,6 +663,8 @@ class FS(object):
             contents = read_file.read()
         return contents
 
+    gettext = _new_name(readtext, "gettext")
+
     def getmeta(self, namespace="standard"):
         # type: (Text) -> Mapping[Text, object]
         """Get meta information regarding a filesystem.
@@ -985,7 +1011,7 @@ class FS(object):
 
     def movedir(self, src_path, dst_path, create=False):
         # type: (Text, Text, bool) -> None
-        """Move contents of directory ``src_path`` to ``dst_path``.
+        """Move directory ``src_path`` to ``dst_path``.
 
         Parameters:
             src_path (str): Path of source directory on the filesystem.
@@ -1083,7 +1109,7 @@ class FS(object):
         with self._lock:
             with self.open(src_path, "rb") as read_file:
                 # FIXME(@althonos): typing complains because open return IO
-                self.setbinfile(dst_path, read_file)  # type: ignore
+                self.upload(dst_path, read_file)  # type: ignore
             self.remove(src_path)
 
     def open(
@@ -1235,7 +1261,7 @@ class FS(object):
             iter_info = itertools.islice(iter_info, start, end)
         return iter_info
 
-    def setbytes(self, path, contents):
+    def writebytes(self, path, contents):
         # type: (Text, bytes) -> None
         # FIXME(@althonos): accept bytearray and memoryview as well ?
         """Copy binary data to a file.
@@ -1253,8 +1279,10 @@ class FS(object):
         with closing(self.open(path, mode="wb")) as write_file:
             write_file.write(contents)
 
-    def setbinfile(self, path, file):
-        # type: (Text, BinaryIO) -> None
+    setbytes = _new_name(writebytes, "setbytes")
+
+    def upload(self, path, file, chunk_size=None, **options):
+        # type: (Text, BinaryIO, Optional[int], **Any) -> None
         """Set a file to the contents of a binary file object.
 
         This method copies bytes from an open binary file to a file on
@@ -1265,21 +1293,28 @@ class FS(object):
             path (str): A path on the filesystem.
             file (io.IOBase): a file object open for reading in
                 binary mode.
+            chunk_size (int, optional): Number of bytes to read at a
+                time, if a simple copy is used, or `None` to use
+                sensible default.
+            **options: Implementation specific options required to open
+                the source file.
 
         Note that the file object ``file`` will *not* be closed by this
         method. Take care to close it after this method completes
         (ideally with a context manager).
 
         Example:
-            >>> with open('myfile.bin') as read_file:
-            ...     my_fs.setbinfile('myfile.bin', read_file)
+            >>> with open('~/movies/starwars.mov', 'rb') as read_file:
+            ...     my_fs.upload('starwars.mov', read_file)
 
         """
         with self._lock:
-            with self.open(path, "wb") as dst_file:
-                tools.copy_file_data(file, dst_file)
+            with self.openbin(path, mode="wb", **options) as dst_file:
+                tools.copy_file_data(file, dst_file, chunk_size=chunk_size)
 
-    def setfile(
+    setbinfile = _new_name(upload, "setbinfile")
+
+    def writefile(
         self,
         path,  # type: Text
         file,  # type: IO
@@ -1299,20 +1334,18 @@ class FS(object):
                 (same as `io.open`).
             newline (str): Newline parameter (same as `io.open`).
 
-        This method will read the contents of a supplied file object,
-        and write to a file on the filesystem. If the destination
-        exists, it will first be truncated.
-
-        If ``encoding`` is supplied, the destination will be opened in
-        text mode.
+        This method is similar to `~FS.upload`, in that it copies data from a
+        file-like object to a resource on the filesystem, but unlike ``upload``,
+        this method also supports creating files in text-mode (if the ``encoding``
+        argument is supplied).
 
         Note that the file object ``file`` will *not* be closed by this
         method. Take care to close it after this method completes
         (ideally with a context manager).
 
         Example:
-            >>> with open('myfile.bin') as read_file:
-            ...     my_fs.setfile('myfile.bin', read_file)
+            >>> with open('myfile.txt') as read_file:
+            ...     my_fs.writefile('myfile.txt', read_file)
 
         """
         mode = "wb" if encoding is None else "wt"
@@ -1349,7 +1382,7 @@ class FS(object):
 
         self.setinfo(path, raw_info)
 
-    def settext(
+    def writetext(
         self,
         path,  # type: Text
         contents,  # type: Text
@@ -1382,6 +1415,8 @@ class FS(object):
         ) as write_file:
             write_file.write(contents)
 
+    settext = _new_name(writetext, "settext")
+
     def touch(self, path):
         # type: (Text) -> None
         """Touch a file on the filesystem.
@@ -1449,7 +1484,7 @@ class FS(object):
                 pass
             else:
                 if len(sys_path) > max_sys_path_length:
-                    _msg = "path too long " "(max {max_chars} characters in sys path)"
+                    _msg = "path too long (max {max_chars} characters in sys path)"
                     msg = _msg.format(max_chars=max_sys_path_length)
                     raise errors.InvalidPath(path, msg=msg)
         path = abspath(normpath(path))
@@ -1570,3 +1605,31 @@ class FS(object):
         from .tree import render
 
         render(self, **kwargs)
+
+    def hash(self, path, name):
+        # type: (Text, Text) -> Text
+        """Get the hash of a file's contents.
+
+        Arguments:
+            path(str): A path on the filesystem.
+            name(str): One of the algorithms supported by the hashlib module, e.g. `"md5"`
+
+        Returns:
+            str: The hex digest of the hash.
+
+        Raises:
+            fs.errors.UnsupportedHash: If the requested hash is not supported.
+
+        """
+        _path = self.validatepath(path)
+        try:
+            hash_object = hashlib.new(name)
+        except ValueError:
+            raise errors.UnsupportedHash("hash '{}' is not supported".format(name))
+        with self.openbin(path) as binary_file:
+            while True:
+                chunk = binary_file.read(1024 * 1024)
+                if not chunk:
+                    break
+                hash_object.update(chunk)
+        return hash_object.hexdigest()
