@@ -27,7 +27,7 @@ except ImportError:
     try:
         from scandir import scandir  # type: ignore
     except ImportError:  # pragma: no cover
-        scandir = None
+        scandir = None  # pragma: no cover
 
 try:
     from os import sendfile
@@ -35,7 +35,7 @@ except ImportError:
     try:
         from sendfile import sendfile  # type: ignore
     except ImportError:
-        sendfile = None
+        sendfile = None  # pragma: no cover
 
 from . import errors
 from .errors import FileExists
@@ -47,7 +47,7 @@ from .path import basename, dirname
 from .permissions import Permissions
 from .error_tools import convert_os_errors
 from .mode import Mode, validate_open_mode
-from .errors import NoURL
+from .errors import FileExpected, NoURL
 
 if False:  # typing.TYPE_CHECKING
     from typing import (
@@ -134,7 +134,7 @@ class OSFS(FS):
                 raise errors.CreateFailed("root path does not exist")
 
         _meta = self._meta = {
-            "case_insensitive": os.path.normcase("Aa") != "aa",
+            "case_insensitive": os.path.normcase("Aa") == "aa",
             "network": False,
             "read_only": False,
             "supports_rename": True,
@@ -151,9 +151,13 @@ class OSFS(FS):
             _meta["invalid_path_chars"] = "\0"
 
             if "PC_PATH_MAX" in os.pathconf_names:
-                _meta["max_sys_path_length"] = os.pathconf(
-                    fsencode(_root_path), os.pathconf_names["PC_PATH_MAX"]
-                )
+                try:
+                    _meta["max_sys_path_length"] = os.pathconf(
+                        fsencode(_root_path), os.pathconf_names["PC_PATH_MAX"]
+                    )
+                except OSError:  # pragma: no cover
+                    # The above fails with nfs mounts on OSX. Go figure.
+                    pass
 
     def __repr__(self):
         # type: () -> str
@@ -245,12 +249,17 @@ class OSFS(FS):
 
     def _gettarget(self, sys_path):
         # type: (Text) -> Optional[Text]
-        try:
-            target = os.readlink(fsencode(sys_path))
-        except OSError:
-            return None
-        else:
-            return target
+        if hasattr(os, "readlink"):
+            try:
+                if _WINDOWS_PLATFORM:  # pragma: no cover
+                    target = os.readlink(sys_path)
+                else:
+                    target = os.readlink(fsencode(sys_path))
+            except OSError:
+                pass
+            else:
+                return target
+        return None
 
     def _make_link_info(self, sys_path):
         # type: (Text) -> Dict[Text, object]
@@ -328,6 +337,8 @@ class OSFS(FS):
         _mode.validate_bin()
         self.check()
         _path = self.validatepath(path)
+        if _path == "/":
+            raise errors.FileExpected(path)
         sys_path = self._to_sys_path(_path)
         with convert_os_errors("openbin", path):
             if six.PY2 and _mode.exclusive:
@@ -454,7 +465,12 @@ class OSFS(FS):
             self.check()
             namespaces = namespaces or ()
             _path = self.validatepath(path)
-            sys_path = self._to_sys_path(_path)
+            if _WINDOWS_PLATFORM:
+                sys_path = os.path.join(
+                    self._root_path, path.lstrip("/").replace("/", os.sep)
+                )
+            else:
+                sys_path = self._to_sys_path(_path)
             with convert_os_errors("scandir", path, directory=True):
                 for dir_entry in scandir(sys_path):
                     info = {
@@ -595,6 +611,8 @@ class OSFS(FS):
         validate_open_mode(mode)
         self.check()
         _path = self.validatepath(path)
+        if _path == "/":
+            raise FileExpected(path)
         sys_path = self._to_sys_path(_path)
         with convert_os_errors("open", path):
             if six.PY2 and _mode.exclusive:
@@ -627,3 +645,17 @@ class OSFS(FS):
                 if accessed is not None or modified is not None:
                     with convert_os_errors("setinfo", path):
                         os.utime(sys_path, (accessed, modified))
+
+    def validatepath(self, path):
+        # type: (Text) -> Text
+        """Check path may be encoded, in addition to usual checks."""
+        try:
+            fsencode(path)
+        except UnicodeEncodeError as error:
+            raise errors.InvalidCharsInPath(
+                path,
+                msg="path '{path}' could not be encoded for the filesystem (check LANG env var); {error}".format(
+                    path=path, error=error
+                ),
+            )
+        return super(OSFS, self).validatepath(path)
